﻿--------------------------------------------------------------
--  Name: Menagerie
--  Author: Markus Treinen
--  Purpose: Swaps certain items when visiting the bank.
-- TODO: Ace3?, ItemID in Datatable als Strings?, ItemInSet ändern?
--------------------------------------------------------------

------------------
-- Global stuff --
------------------

-- create an instance of AceLibrary
Menagerie = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceEvent-2.0")

-- Library support
local L = AceLibrary("AceLocale-2.2"):new("Menagerie")

local Data

--------------------
-- Event handlers --
--------------------

-- set defaults and register SavedVariables
function Menagerie:OnInitialize()
	self:RegisterDB("MenagerieDB")
	self:RegisterDefaults("profile", {
		swapMiniPets = true,
		swapMounts = true,
		swapTabards = false,	-- disabled by default
		Blacklist = {},
	})
	self:RegisterChatCommand({"/menagerie", "/mg"}, {
		type = "group",
		args = {
			minipets = {
				type = "toggle",
				order = 100,
				name = L["SWAPMINIPETS"],
				desc = L["SWAPMINIPETS_DESC"],
				get =	function()
							return self.db.profile.swapMiniPets
						end,
				set =	function(v)
							self.db.profile.swapMiniPets = v
							self:InitCategories()
						end,
			},
			mounts = {
				type = "toggle",
				order = 200,
				name = L["SWAPMOUNTS"],
				desc = L["SWAPMOUNTS_DESC"],
				get =	function()
							return self.db.profile.swapMounts
						end,
				set =	function(v)
							self.db.profile.swapMounts = v
							self:InitCategories()
						end,
			},
			tabards = {
				type = "toggle",
				order = 300,
				name = L["SWAPTABARDS"],
				desc = L["SWAPTABARDS_DESC"],
				get =	function()
							return self.db.profile.swapTabards
						end,
				set =	function(v)
							self.db.profile.swapTabards = v
							self:InitCategories()
						end,
			},
			blacklist = {
				type = "group",
				order = 400,
				name = L["BLACKLIST"],
				desc = L["BLACKLIST_DESC"],
				args = {
					add = {
						type = "text",
						order = 100,
						name = L["ADD_BLACKLIST"],
						desc = L["ADD_BLACKLIST_DESC"],
						usage = L["ADD_BLACKLIST_USAGE"],
						get = false,
						set =	function(item)
									self:AddBlacklistItem(item)
								end,
					},
					del = {
						type = "text",
						order = 200,
						name = L["DEL_BLACKLIST"],
						desc = L["DEL_BLACKLIST_DESC"],
						usage = L["DEL_BLACKLIST_USAGE"],
						get = false,
						set =	function(item)
									self:DelBlacklistItem(item)
								end,
					},
					clear = {
						type = "execute",
						order = 300,
						name = L["CLEAR_BLACKLIST"],
						desc = L["CLEAR_BLACKLIST_DESC"],
						func =	function()
									self:ClearBlacklist()
								end,
					},
					list = {
						type = "execute",
						order = 400,
						name = L["LIST_BLACKLIST"],
						desc = L["LIST_BLACKLIST_DESC"],
						func =	function()
									self:ListBlacklist()
								end,
					},
				},
			},
		},
	})
end

-- init datastructres and register events when enabled
function Menagerie:OnEnable()
	self:ClearData()
	self:ClearItems()
	self:InitIDs()
	self:InitCategories()

	self:RegisterEvent("BANKFRAME_OPENED")
end

-- erase datastructures when disabled
function Menagerie:OnDisable()
	self:ClearData()
end

-- do the necessary magic when opening the bank
function Menagerie:BANKFRAME_OPENED()
	-- search all given container classes (bags and bank) for valid candidates for swapping
	self:FillItems()
	
	-- do the swap between the relevant container classes (bags and bank)
	self:SwapItemsUsable("Bags", "Bank")
	
	-- do the swap between the relevant container class (bank) and the worn clothes
	self:SwapItemsWorn("Bank")
	
	-- wipe data after our work is done
	self:ClearItems()
end

------------------------------
-- Initialization functions --
------------------------------

-- Clears and initializes the main datastructure Data
function Menagerie:ClearData()
	Data = {}
end

-- Clears and initializes the Data.Items-table
function Menagerie:ClearItems()
	Data.Items = {
		Usable = {},
		Worn = {},
	}
end

-- Clears and fills the Data.IDs-table
function Menagerie:InitIDs()
	-- those IDs are the defaults and bag 0 is hardcoded
	Data.IDs = {
		Bags = { 0 },
		Bank = { BANK_CONTAINER }
	}
	
	-- bag slots are numbered from 0 (bag) to NUM_BAG_SLOTS
	-- (additional bags)
	for i = 1, NUM_BAG_SLOTS do
		table.insert(Data.IDs.Bags, i)
	end
	
	-- bank slots begin after the last bag slot and there are
	-- NUM_BANKBAGSLOTS slots
	for i = NUM_BAG_SLOTS+1, NUM_BAG_SLOTS+NUM_BANKBAGSLOTS do
		table.insert(Data.IDs.Bank, i)
	end
end

-- Rebuilds the table containing enabled categories
function Menagerie:InitCategories()
	local profile = self.db.profile

	Data.Categories = {
		Usable = {},
		Worn = {},
	}
	
	if profile.swapMounts then
		table.insert(Data.Categories.Usable, "Mount")
	end
	if profile.swapMiniPets then
		table.insert(Data.Categories.Usable, "Minipet")
	end
	if profile.swapTabards then
		table.insert(Data.Categories.Worn, "Tabard")
	end
end

--------------------
-- Main functions --
--------------------

-- Fills the Data-table with existing slot/bag-positions to the
-- corresponding enabled categories
function Menagerie:FillItems()
	-- iterate all given inventories (i.e. bags, bank)
	for inventory in pairs(Data.IDs) do

		-- iterate all bags of that inventory
		for _, bagid in pairs(Data.IDs[inventory]) do

			-- iterate all slots of that bag
			for slotid = 1, GetContainerNumSlots(bagid) do
				local itemlink = GetContainerItemLink(bagid, slotid)
			
				-- if the slot contains an item and it is not
				-- blacklisted...
				if itemlink and not self:IsBlacklisted(itemlink) then

					-- do an item search for all Categories
					for base in pairs(Data.Categories) do
					
						-- ...then browse all given and enabled categories for that item
						for _, category in pairs(Data.Categories[base]) do

							-- determine the set in which the item is in
							local set = self:ItemInSet(itemlink, category)

							-- if a set containing that item is found...
							if set then

								-- create category table, if it doesn't
								-- exist
								if not Data.Items[base][set] then
									Data.Items[base][set] = {}
								end

								-- create inventory table under the
								-- category table, if it doesn't exist
								if not Data.Items[base][set][inventory] then
									Data.Items[base][set][inventory] = {}
								end

								-- ...then create an entry in our huge table
								table.insert(Data.Items[base][set][inventory], {
									bag = bagid,
									slot = slotid,
								})
								
							end -- if set

						end -- for Data.Categories

					end -- for base

				end -- if itemlink and not blacklisted

			end -- for slots

		end -- for bags

	end -- for i

end

-- Swaps the items between inv1 and inv2
function Menagerie:SwapItemsUsable(inv1, inv2)
	-- iterate every category of items
	for _, items in pairs(Data.Items.Usable) do
		local itemsinv1 = items[inv1]
		local itemsinv2 = items[inv2]
		
		-- do we have something to swap?
		if itemsinv1 and itemsinv2 then
			local num1 = # itemsinv1
			local num2 = # itemsinv2

			-- only do the lowest number of swaps, more swaps don't
			-- make sense
			if num1 > num2 then
				num1, num2 = num2, num1
				itemsinv1, itemsinv2 = itemsinv2, itemsinv1
			end

			-- iterate over every item of the category in inventory 1
			for orig = 1, num1 do
				-- and pick a random counterpart from inventory 2
				local swap = math.random(num2)

				-- do the swap
				PickupContainerItem(itemsinv1[orig].bag, itemsinv1[orig].slot)
				PickupContainerItem(itemsinv2[swap].bag, itemsinv2[swap].slot)

				-- remove the swapped item from inventory 2, so it
				-- doesn't get swapped again
				table.remove(itemsinv2, swap)

				-- decrease the number of still swappable items in
				-- inventory 2
				num2 = num2 - 1
			end -- for num1
		end -- if itemsinv1 and itemsinv2
	end -- for Data.Items.Usable
end

-- Swaps the found items between inv1 and worn clothes
function Menagerie:SwapItemsWorn(inv1)
	-- iterate every category of items
	for _, items in pairs(Data.Items.Worn) do
		local itemsinv1 = items[inv1]
		
		-- do we have something to swap?
		if itemsinv1 then
			local num1 = # itemsinv1
			
			-- pick a random item
			local swap = math.random(num1)
			
			-- do the swap
			PickupContainerItem(itemsinv1[swap].bag, itemsinv1[swap].slot)
			AutoEquipCursorItem()
		end
	end
end

--------------------------
-- Blacklist management --
--------------------------

-- Adds an ItemID to the blacklist
function Menagerie:AddBlacklistItem(item)
	local itemid, itemlink = self:ItemInfo(item)

	if itemid then
		if not self:IsBlacklisted(itemid) then
			self.db.profile.Blacklist[itemid] = true
			self:Print(string.format(L["BLACKLIST_ADDED"], itemlink))
		else
			self:Print(string.format(L["BLACKLIST_ALREADY"], itemlink))
		end
	else
		self:Print(string.format(L["ERROR_NOTCACHED_OR_INVALID"], item))
	end
end

-- Removes an ItemID from the blacklist
function Menagerie:DelBlacklistItem(item)
	local itemid, itemlink = self:ItemInfo(item)

	if itemid then
		if self:IsBlacklisted(itemid) then
			self.db.profile.Blacklist[itemid] = nil
			self:Print(string.format(L["BLACKLIST_REMOVED"], itemlink))
		else
			self:Print(string.format(L["BLACKLIST_MISSING"], itemlink))
		end
	else
		self:Print(string.format(L["ERROR_NOTCACHED_OR_INVALID"], item))
	end
end

-- Output the whole blacklist
function Menagerie:ListBlacklist()
	-- a little trick to test if a table is empty
	if next(self.db.profile.Blacklist) ~= nil then
		self:Print(L["BLACKLIST_LISTING"])
	
		for item, status in pairs(self.db.profile.Blacklist) do
			local itemid, itemlink = self:ItemInfo(item)
		
			if itemid then
				self:Print("    " .. itemlink .. " (" .. itemid .. ")")
			else
				self:Print("    (" .. L["NOTCACHED"] .. ") " .. item)
			end
		end
	else
		self:Print(L["BLACKLIST_EMPTY"])
	end
end

-- Nuke the blacklist
function Menagerie:ClearBlacklist()
	self.db.profile.Blacklist = {}
	self:Print(L["BLACKLIST_CLEARED"])
end

-- Returns whether the item is blacklisted or not
function Menagerie:IsBlacklisted(item)
	local itemid = self:ItemID(item)
	return (itemid ~= nil) and (self.db.profile.Blacklist[itemid] == true)
end

-----------------------
-- Itemset functions --
-----------------------

-- Returns the set name which item belongs to
function Menagerie:ItemInSet(item, set)
	local itemid = self:ItemID(item)
	local safeset = set:gsub("([%.%(%)%%%+%-%*%?%[%]%^%$])", "%%%1")
	
	for k, v in pairs(MenagerieData) do
		-- not quite perfect, but what the hell... ;)
		-- LUA's pattern matching seems to be a bit limited (Perl 4tw!)
		if k:find("^" .. safeset) then
			if v[itemid] then
				return k
			end
		end
	end
end

---------------------
-- Tool functions ---
---------------------

-- Returns ItemID and ItemLink either from an ItemLink, ItemID or
-- ItemName (nil if not in cache)
function Menagerie:ItemInfo(item)
	local _, itemlink = GetItemInfo(item)
	return itemlink and self:ItemID(itemlink), itemlink
end

-- Returns ItemID from ItemLink or ItemID (nil if not valid)
function Menagerie:ItemID(item)
	return tonumber(item) or tonumber(item:match("item:(%d+)"))
end
