#!/usr/bin/perl -T

use strict;
use warnings;
use LWP::Simple;
use Text::Balanced qw(
	extract_bracketed
	extract_delimited
);

# ---
my $url_minipets = "http://www.wowhead.com/?items=15.2";
my $url_minipets_snowball = "http://www.wowhead.com/?items=15.2&filter=cr=107;crs=0;crv=Requires+a+Snowball+to+summon.";

my $url_mounts_ground_fast = "http://www.wowhead.com/?items=15.-4&filter=qu=4:5:6;cr=111;crs=2;crv=75";
my $url_mounts_ground_slow = "http://www.wowhead.com/?items=15.-4&filter=qu=3;cr=111;crs=2;crv=75";

my $url_mounts_flying_fast = "http://www.wowhead.com/?items=15.-7&filter=qu=4:5:6";
my $url_mounts_flying_slow = "http://www.wowhead.com/?items=15.-7&filter=qu=3";

my $url_mounts_aq40 = "http://www.wowhead.com/?items=15.-4&filter=qu=3;cr=111:107;crs=2:0;crv=75:within+the+Temple+of+Ahn%27Qiraj";

my $url_tabards = "http://www.wowhead.com/?items=4&filter=sl=19";

my $data_file = "MenagerieData.lua";
my $table_name = "MenagerieData";

my $debug = 1;		# set higher for more debug output (current max = 3)
# ---

my %Data;

### Output something based on $debug flag
sub Print {
	my $level = shift;

	print @_ if ($level <= $debug);
}

### Parse Wowhead-page specified via $url.
### Put all items found in the big Data hash for $category.
### If $category_delete is specified, all items found for $category will be
### erased from $category_delete. Wowhead has no way (yet?) to exclude items
### with a specific effect text.
sub parse_page {
	my ($url, $category, $category_delete) = @_;

	Print(1, "Parsing '$category'");
	Print(2, " with deleting from '$category_delete'")	if ($category_delete);
	Print(2, " from URL '$url'");
	Print(1, "\n");

	die "No category specified for $url!\n"		unless $category;

	Print(3, "getting webpage\n");
	my $content = get($url);
	die "Error fetching page $url!\n"		unless $content;

	Print(3, "parsing Listview\n");
	my $listview = extract_bracketed($content, "(\"')", "(?s).*?new Listview\\s*");
	Print(3, "-> $listview\n");

	die "No items found on page $url!\n"		unless $listview;

	Print(3, "parsing data\n");
	my $items = extract_bracketed($listview, "[\"']", "(?s).*?data:\\s*");
	Print(3, "-> $items\n");

	die "No item data found on page $url!\n"	unless $items;

	Print(3, "iterating items\n");
	my ($item, $rest);
	while (($item, $rest) = extract_bracketed($items, "{\"'}", "(?s)[^\\{]") and $item) {
		Print(3, "extract ID");
		my ($id) = $item =~ /id:(\d+)/;
		Print(3, " '$id'\n");

		Print(3, "parsing name");
		my $name = extract_delimited($item, "'\"", "(?s).*?name:\\s*");
		Print(3, " '$name'\n");

		die "No itemid or item name found!\n"	unless $id and $name;

		Print(3, "modifying name '$name'");
		$name =~ s/\\//g;
		$name =~ s/^'\d//;
		$name =~ s/'$//;
		Print(3, " -> '$name'\n");

		Print(3, "inserting in native Data structure\n");
		$Data{$category}{$id}{Name} = $name;
		$Data{$category}{$id}{Value} = 1;

		if ($category_delete) {
			Print(3, "deleting from '$category_delete'\n");
			delete $Data{$category_delete}{$id};
		}
	}
}
 
### Generates a nice looking LUA-table from the big Data hash.
sub generate_data {
	my $data = "";

	Print(1, "Generating Data");
	Print(2, " with table name '$table_name'");
	Print(1, "\n");

	$data .= "$table_name = {\r\n";

	Print(3, "iterating categories\n");
	foreach my $category (sort(keys(%Data))) {
		Print(3, "category '$category'\n");

		$data .= "\t[\"$category\"] = {\r\n";

		Print(3, "iterating IDs\n");
		foreach my $id (sort({$a <=> $b} keys(%{$Data{$category}}))) {
			Print(3, "ID '$id'\n");

			$data .= "\t\t[$id] = true,\t-- $Data{$category}{$id}{Name}\r\n";
		}
		$data .= "\t},\r\n";
	}

	$data .= "}\r\n";

	return $data;
}

### Writes the generate_data() to the $data_file.
sub write_datafile {
	my ($file) = @_;

	Print(1, "Writing Datafile");
	Print(2, " '$file'");
	Print(1, "\n");

	Print(3, "opening for writing\n");
	unless (open(DATA, ">", $file)) {
		die "Couldn't write datafile $file: $!\n";
	}

	Print(3, "writing data\n");
	binmode DATA;
	print DATA generate_data();

	Print(3, "closing file\n");
	close DATA;
}

### Main procedure
sub main {
	parse_page($url_minipets, "Minipet.Normal");
	parse_page($url_minipets_snowball, "Minipet.Snowball", "Minipet.Normal");

	# ---

	parse_page($url_mounts_ground_slow, "Mount.Ground.Slow");
	parse_page($url_mounts_ground_fast, "Mount.Ground.Fast");

	parse_page($url_mounts_aq40, "Mount.Ground.AQ40", "Mount.Ground.Slow");

	parse_page($url_mounts_flying_slow, "Mount.Flying.Slow");
	parse_page($url_mounts_flying_fast, "Mount.Flying.Fast");

	# ---

	parse_page($url_tabards, "Tabard");

	# ---

	write_datafile($data_file);
}

main();
